# -*- coding: utf-8 -*-
"""
Created on Fri Feb  3 15:18:39 2023

@author: Mohammed
"""

import numpy as np 
import math 

def Gauss_distr(X, sigma=0.1,): 
    ##  g(x)=1/(sigma*sqrt(2*pi))*{exp(-(x-b)^2/(2*sigma^2))}
    norm = (sigma*np.sqrt(2*np.pi))**-1 
    
    if np.ndim(X) ==1: 
        x=np.arange(X[0], X[-1],0.05)
        Y = np.zeros((len(x)))
        #Dressing with gaussian function
        for J in X:
            y =norm * np.exp(-1*(x-J)**2/(2*sigma**2))
            Y +=y 
        return x,Y 
    elif np.ndim(X) ==2: 
        x=np.arange(X.min(), X.max(),0.05)
        YY = np.zeros((len(x), X.shape[1])) 
        for i in range(X.shape[1]): 
            Y = np.zeros((len(x)))
        #Dressing with gaussian function
            for J in X[:,i]:
                y =norm * np.exp(-1*(x-J)**2/(2*sigma**2))
                Y +=y 
            YY[:,i] = Y  
        return x, YY 
        

def Dress_abs(X,OS, unit=None, linewidth=0.4,): 
    """ 
    Absorption calculated using following equation. 
    Ei = {sqrt(pi)*e^2*N / (1000*ln(10).c^2*me)}* {fi/sigma} * exp[-{(x-xi) / sigma}^2 ] 
    electron charge: e = 4.803204×10-10 esu (cm3/2.g1/2.s-1) 
    electron mass: me = 9.10938×10-28 g
    Avogrado number: N = 6.022*10**23 mol-1 
    speed of light: c=29979245800.0 cm·s-1 
    sigma in eV unit, and converted to nm unit 
""" 
    e = 4.803204*10**-10
    me = 9.10938*10**-28 
    c = 29979245800.0  # cm·s-1 
    N = 6.022*10**23  
    
    sigma_cm =  10**-7*1240/linewidth #linewidth in ev unit 
         
    norm = np.sqrt(np.pi)*N*e**2 / (1000*np.log(10)*me*(1/sigma_cm)*c**2)
    
    ev2nm = lambda a : 1240/a 
    
    if unit: 
        if unit.lower() == 'ev': 
            x =np.arange(min(X[0],0.05), max(X[-1],12),0.005) 
        elif unit.lower() == 'nm': 
            x = np.logspace(np.log2(max(3000,X[0])), np.log2(min(100,X[-1])), num=2000 , base = 2, endpoint=False)
    else: 
        if X.max() > 50:
            unit='nm'
            print('Absorption spectra is dressing as nm unit')
            x = np.logspace(np.log2(max(3000,X[0])), np.log2(min(100,X[-1])), num=2000 , base = 2, endpoint=False)
        else: 
            unit = 'ev'
            x =np.arange(min(X[0],0.1), max(X[-1],12),0.005)

    Y = np.zeros((len(x)))
    #Dressing with gaussian function
    if unit == 'ev': 
        for j in range(len(X)):
            y =norm * OS[j] *  np.exp(-1*((x-X[j])/linewidth)**2) 
            Y +=y 
        return x,Y 
    elif unit=='nm': 
        for j in range(len(X)):
            y =norm * OS[j] *  np.exp(-1*((1/x-1/X[j])/(1/ev2nm(linewidth)))**2)   
            Y +=y 
        return x,Y


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 19 11:25:53 2022

@author: Mohammed A Jabed Jabed.abu@gmail.com 
"""


#sns.set(context='notebook', style='white',
#        color_codes=True, rc={'legend.frameon': False})
import matplotlib.pyplot as plt
import math 
import numpy as np 

plt.rcParams["font.family"] = 'Arial'
plt.rcParams.update({'font.size': 20})
Color = plt.rcParams['axes.prop_cycle'].by_key()['color']
#Color = ['red', 'skyblue', 'green','gray','orange']

def plot_line(x,y, xlim=None, ylim=None, ls=None, lw=None, colors = None, xticks=None, yticks=None, 
              xticklabels=None, yticklabels=None,xlabel=None, ylabel=None, figsize=(7,5), name=None):  

    argnan = lambda a,b : a if a != None else b 
    
    fig=plt.figure(figsize=figsize) 
    ax= fig.add_subplot(1,1,1) 

    ax.plot(x,y,color=argnan(colors,'b'),lw=argnan(lw,2),ls=argnan(ls,'-')) 
    if xlim: 
        ax.set_xlim(xlim[0],xlim[1]) 
    if ylim: 
        try: 
            ax.set_ylim(ylim[0],ylim[1])
        except: 
            ax.set_ylim(bottom=0) 
    if yticks: 
        ax.set_yticks(yticks) 
    if xticks: 
        ax.set_xticks(xticks) 
    if yticklabels: 
        ax.set_yticklabels(yticklabels) 
    if xticklabels: 
        ax.set_xticklabels(xticklabels) 
    if xlabel: 
        ax.set_xlabel(xlabel) 
    if ylabel: 
        ax.set_ylabel(ylabel) 
    if name: 
        plt.tight_layout()
        plt.savefig(name,dpi=330)         
    else: 
        return ax 


def plot_waterfall(x,z,xlim=[-3,3],ylim=None, gamma = 45, theta=45, alpha=0.4,
                   figsize=(7,5),name=None, ls=None, lw=None, color = None, xlabel=None, ylabel=None ): 
    print('This is experimental work, outcome may not produce expected quality')
    n_prec =  0 if math.floor(math.log10(z.max() - z.min())) > 1 else 2-  math.floor(math.log10(z.max() - z.min()))
    z_shift = z.min() 
    if z_shift<0: 
        z = z-z_shift  
    gamma= math.radians(gamma) 
    theta = math.radians(theta) 
    
    Rx = np.array([[1, 0,0],
                   [0, math.cos(theta), math.sin(theta)],
                   [0,-math.sin(theta),math.cos(theta)]])   
    Rz = np.array([[math.cos(gamma), math.sin(gamma), 0],
                   [-math.sin(gamma),math.cos(gamma),0],
                   [0,0,1]])  
    # Rotation matrix around x and z axis 
    Rxz = np.array([[math.cos(gamma),math.cos(theta)*math.sin(gamma), math.sin(theta)*math.sin(gamma)],
                    [-math.sin(gamma), math.cos(theta)*math.cos(gamma),math.sin(theta)*math.cos(gamma)],
                    [0,-math.sin(theta),math.cos(theta)]])


    fig = plt.figure(figsize=(7,5)) 
    ax = fig.add_subplot(1,1,1)  
    
    x_range = xlim[1] - xlim[0] 
    if ylim:
        yz_ratio = (ylim[1]-ylim[0])*0.2
    else: 
        yz_ratio = (z.max()-z.min())*0.2
    
    if z.shape[1]*x_range*0.02 > x_range*0.25: 
        yx_ratio = 0.25/z.shape[1] 
    else: 
        yx_ratio = 0.02      
    print(yx_ratio)
    for i in np.arange(0,z.shape[1],1): 
        ay = [0,i*x_range*yx_ratio,0]
        ay_trans = np.matmul(Rxz,ay) 
        X = x + np.abs(ay_trans[0]) 
        Y = z[:,i] + np.abs(ay_trans[2]) * yz_ratio
        ax.fill_between(X, Y, 0, color='white',edgecolor='w', zorder=-i, alpha=alpha)
        if color: 
            ax.plot(X,Y,zorder=-i,color=color ) 
        else: 
            ax.plot(X,Y,zorder=-i)
    #   print(ay_trans)
        
    ax.set_ylim(bottom = z.min())
    ylim = ax.get_ylim()

    X2 = xlim[0] + np.abs(ay_trans[0] ) 
    Y2 = np.abs(ay_trans[2]) * yz_ratio 

    ax.set_xlim(xlim[0],X2+x_range)
    axlw = ax.spines['bottom'].get_linewidth() 

    ax.plot([xlim[0],X2], [z.min(),z.min()+Y2],color='k', zorder=51)  
    ax.plot([xlim[1],X2+x_range], [z.min(),z.min()+Y2],color='k', zorder=51)  

    ax.plot([xlim[0],X2], [ylim[1],ylim[1]+Y2],  lw = axlw, color='k', zorder=51)  
    ax.plot([xlim[1],X2+x_range], [ylim[1],ylim[1]+Y2], lw = axlw, color='k', zorder=51)   

    ax.vlines(x=X2,ymin=Y2, ymax =ylim[1] + Y2,  lw = axlw, color='k', zorder = -2)  
    ax.vlines(x=xlim[1] , ymin=z.min(), ymax =ylim[1],  lw = axlw, color='k', zorder = 10)  
    #print( X2)
    ax.hlines(y=Y2, xmin = X2 ,xmax = x_range + X2,  lw = axlw, color = 'k', zorder = -5)  
    ax.hlines(y=ylim[1], xmin= xlim[0], xmax = xlim[1],  lw = axlw, color='k', zorder = -10)  

    ax.set_ylim(0,ylim[1]+Y2)
 
    ax.spines['bottom'].set_bounds(xlim[0], xlim[1])
    ax.spines['top'].set_bounds(X2, X2+x_range) #xlim[1]+np.abs(ay_trans[0]))
    ax.spines['right'].set_bounds(Y2, ylim[1] + Y2 )  
    ax.spines['left'].set_bounds(z.min(), ylim[1] )#+ Y2 )   

    new_ticks = ax.get_yticks()+z_shift 
    print(new_ticks)
    ax.set_yticklabels([f"{j:.{n_prec}f}" for j in new_ticks])
    
    if xlabel: 
        ax.set_xlabel(xlabel) 
    if ylabel: 
        ax.set_ylabel(ylabel) 
    if name: 
        #plt.tight_layout()
        plt.savefig(name,dpi=630)         
    else: 
        return ax 
        



import math, os.path, sys, time
from datetime import datetime, timedelta
from glob import glob
from argparse import ArgumentParser
import numpy as np 
import os 

try : 
    from read_outcar import * 
    from Plot import * 
    from fileconversion import * 
    from mathstats import * 
    
except: 
    from .read_outcar import * 
    from .Plot import * 
    from .fileconversion import * 
    from .mathstats import * 
    
class Warning: 
    def file_not_exist( name): 
        if  not  os.path.isfile(name): 
            print('warning! file not found - %s' %name) 
            exit()
        else : pass 
    def VaspOrGaus(fname, vasp=False, gaussian=False): 
        ff = open(fname)
        ini_line = ff.readline() 
        if not 'Gaussian' in ini_line:
            if gaussian == True: 
                exit('%s is not a gaussian output log file' %fname) 
        elif not 'vasp' in ini_line: 
            if vasp == True : 
                exit('%s is not a VASP OUTPUT type file'%fname) 
        else: pass  

    
def main(): 
    parser = ArgumentParser() 
    parser.add_argument("-job", dest="job", action="store", default=None, type=str.lower,
                        help="task to perform")
    parser.add_argument("--file", dest="fname", default=None, type=str, 
                        help = 'input file name'),
    parser.add_argument("--fout", dest="fout", default=None, type=str, 
                        help = 'output file name'),
    parser.add_argument("--ftype", dest="ftype", default="xyz", type=str, 
                        help = 'output file type')
    parser.add_argument("--select", dest="select", default=False, type=lambda x: True if x=='True' else False,  
                        help = 'output file type')

    parser.add_argument("--direct", dest="direct", default=True, type=lambda x: False if x=='False' else True, 
                        help = 'output file type')
    parser.add_argument("--bands", dest="bands", default=100, type=int, 
                        help = 'No of states to read, up-below homo')
    parser.add_argument("--frames", dest="frames", default='final', type=str, 
                        help = 'No of MD frame to read, e.g. final, initial, all, -n, n')
    parser.add_argument("--nthkpoint", dest="nthkpoint", default=1, type=int, 
                        help = 'Read nth KPOINT data')
    parser.add_argument("--coord", dest="coord", default='cart', type=str, 
                        help = 'Output coordinate types')

    parser.add_argument("--plot", dest="plot",  type=lambda x: True if x=='True' else x, default=False, 
                        help="Plot output")
    parser.add_argument("--sigma", dest='sigma',default=0.1, type=float,
                        help = 'Linewidth for the Gaussian broadening, default 0.1 eV')
    parser.add_argument("--shiftfermi", dest="shiftfermi", action='store_true', default=False, 
                        help = 'Plot shift to Fermi energies')
    parser.add_argument("--xlim", dest="xlim",  nargs='+', type=float, default=None, 
                        help = 'Lower and upper limit of the X axis')
    parser.add_argument("--alpha", dest="alpha",  type=float, default=0.4, 
                        help = 'Set transpaency of a matplotlib graph') 
    parser.add_argument("--theta", dest="theta", type=float, default=45, 
                        help = 'X-axis rotation for 3D view of waterfall graphs')    
    parser.add_argument("--gamma", dest="gamma", type=float, default=20, 
                        help = 'Z-axis rotation for 3D view of waterfall graphs')    
    parser.add_argument("--unit", dest="unit", type=str.lower, default='ev', 
                            help = 'Absorption spectra in eV or nm unit?, default in eV unit') 
    parser.add_argument("--figsize", dest="figsize",  nargs='+', type=int, default=(7,5), 
                            help = 'Plot figure size, integer list, default (7,5)')    
    
    
    (options, args) = parser.parse_known_args() 
    #print(options)
   #print( args)
    
    if options.job == 'vasp2gaus': 
        if options.fname is None: 
            options.fname = input('What is the POSCAR file name? :').lstrip().rstrip() 
        fname = options.fname
        Warning.file_not_exist(fname)  
        
        if options.fout is None: 
            options.fout = os.path.splitext(options.fname)[0]
        name_out = os.path.splitext(options.fout)[0]  
        
        if options.ftype =='gaus': 
            Gaus_out = True 
            if os.path.isfile('%s.com'%name_out):
                print ('%s.com is exist'%name_out) 
                name_out_ = input('What should be the output file name?:').rstrip().lstrip().rstrip('.com')  
                if not len(name_out_)==0: 
                    name_out= name_out_     
                if os.path.isfile('%s.com'%name_out):  # overwritting the file if no name is given 
                    print('Overwriting the file %s.com') # name_out = name_out+'_'+''.join(random.choices(string.ascii_lowercase, k=5)) 
                    os.remove('%s.com'%name_out) 
        if options.ftype =='xyz': 
            Gaus_out = False
            if os.path.isfile('%s.xyz'%name_out):
                print ('%s.xyz is exist'%name_out) 
                name_out_ = input('What should be the output file name?:').rstrip().lstrip().rstrip('.xyz')  
                if not len(name_out_)==0: 
                    name_out= name_out_     
                if os.path.isfile('%s.xyz'%name_out): 
                    os.remove('%s.xyz'%name_out) 
                    print('Overwriting the file %s.xyz'%name_out) 
   
        contcar2com(fname, name_out,Gaus_out=Gaus_out) 
    
    elif options.job == 'gaus2poscar': 
        fname = options.fname 
        
        Warning.file_exist(fname)  
        
        file = open(fname,'r') 
        lines = file.read() 
        if not 'tv' in lines.lower() :
            print('Warning! Periodic cell vectors are not found in file %s' %fname)   
            exit('Add periodic cell in the com file')
        file.close()  
        
        name_out = os.path.splitext(options.fout)[0] 
        
        com2poscar(fname, name_out=name_out, select=options.select, direct = options.direct)  

    elif options.job == 'read_bands':
        fname = options.fname 
        nbands = options.bands 
        frames = options.frames
        nthkpoints = options.nthkpoint -1 
        
        #Translate frame argument in int variables 
        get_frame =  lambda a, k, nf : a[k,-1,:,:] if nf == 'final' \
        else ( a[k,0,:,:] if nf == 'initial' \
              else (a[k,:,:,:] if  nf ==  'all' 
                    else (a[k,:int(nf),:,:] if nf.isdigit() 
                          else (a[k,int(nf):,:,:] if nf.lstrip('-').isdigit()
                              else a ))))   
        
        outcar = read_outcar(filename=fname)
        Bands_all = outcar.band_energies()   # Output dimension is K points x frames x bands x 3 
        
        Bands = get_frame(Bands_all,nthkpoints, frames) 
        
        Fermi = outcar.fermi_energies()
        Fermi_frames = Fermi[-1,0] if frames =='final' else (Fermi[0,0] if frames=='initial' 
                                              else(Fermi[:,0] if frames=='all'
                                                   else (Fermi[int(frames),0] if frames.isdigit() 
                                                         else(Fermi[int(frames),0] if frames.lstrip('-').isdigit() 
                                                              else Fermi[:,0] ))))
        print(Fermi[0], 'Fermi zerooooooo')
        nhomo  = int(Bands[0,:,0][Bands[0,:,1] < Fermi_frames[0]][-1])  
        nbands = nbands if nbands < nhomo else nhomo-1 
        
        if options.plot == True:  
            print(Bands_all.shape)
            x = Bands_all[nthkpoints,-1,:,1] .astype(float)
           
            if options.shiftfermi: 
                x = x - outcar.fermi_energies()[-1,0]  
            if options.xlim: 
                xlim=options.xlim 
                #print(xlim)
                x = x[(x>xlim[0]) & (x<xlim[1]) ] 
            sigma = options.sigma
            x,y = Gauss_distr(x, sigma=sigma )  
            if not options.fout : 
                name = fname
            else: 
                name ='DOS_%s'% options.fout 
                
            plot_line(x,y, name = name, xlim=xlim, ylim=(0,),ls=None, lw=None, colors ='g', xticks=None, yticks=None, 
                          xticklabels=None, yticklabels=None,xlabel=r'E-$E_{fermi}$', ylabel=None)  
            np.savetxt('%s.dat'%name, np.c_[x,y],fmt='%0.4f  %0.4f' )    
        # Data write in a file options.fout, else print
    
        elif options.plot=='dos_all': 
            if not options.fout : 
                name = fname
            else: 
                name = options.fout
            if options.xlim: 
                xlim=options.xlim 
            else: 
                xlim=[Fermi_frames[0]-3,Fermi_frames[0]+3]   
            Bands_for_dos = Bands [:,:,1] 

            if options.shiftfermi: 
                Bands_for_dos =Bands_for_dos - Fermi_frames[:,None]
                xlim = xlim - Fermi_frames[0]
            print(xlim)
           # print(Bands_for_dos[0,:] >  ) 
            if options.xlim: 
                 xlim=options.xlim 
                 #print(xlim)
                 Bands_for_dos = Bands_for_dos.T[(Bands_for_dos[0,:] >xlim[0]) & (Bands_for_dos[0,:]<xlim[1]) ]
 #                Bands_for_dos = Bands_for_dos
            sigma = options.sigma            
            x,y = Gauss_distr(Bands_for_dos, sigma=sigma) 

            plot_waterfall(x,y,xlim=xlim,ylim=None, gamma = options.gamma, theta=options.theta, alpha=options.alpha,
                           color=None,name='DOS_all_%s'%name)
            np.savetxt('bands_all_%s.dat'%name,Bands_for_dos,fmt='%0.5f') 
            np.savetxt('Dress_dos_all_%s.dat'%name, np.c_[x,y], fmt='%0.4f') 

        if options.fout != None: 
            Bands = Bands[:,nhomo-nbands-1:nhomo+nbands-1,:] 
            if Bands.ndim <3: #it indicates one frame only 
                with open(options.fout,'w') as f: 
                    np.savetxt(f,Bands, fmt='%-4.i   %-10.6f    %-10.6f')
                f.close() 
            elif Bands.ndim ==3: 
                with open(options.fout,'w') as f: 
                    for j in range(Bands.shape[0]): 
                        f.write('Frame %i \n'%(j+1)) 
                        np.savetxt(f,Bands[j,:,:], fmt='%-4.i   %-10.6f    %-10.6f') 
                        f.write('\n')
            else: 
                print('Can not find a way to write in a file') 
                print(Bands) 
        else: 
            Bands = Bands[:,nhomo-nbands-1:nhomo+nbands-1,:]
            print(Bands) 
            
    elif options.job == 'get_geom': #read geometry/geomteries from optimization steps or MD trajectory from VASP calculations  
        fname = options.fname 
        Warning.file_not_exist(fname)
        Warning.VaspOrGaus(fname, vasp=True) 
        fout = options.fout   
        coord = options.coord 
        frames = options.frames
        
        get_frame =  lambda a, n : a[-1,:,:] if n == 'final' \
        else ( a[0,:,:] if n == 'initial' \
              else (a[:,:,:] if  n ==  'all' 
                    else (a[:int(n),:,:] if n.isdigit() 
                          else (a[int(n):,:,:] if n.lstrip('-').isdigit()
                              else a )))) 
        
        outcar = read_outcar(filename=fname)
        Trajectory = outcar.md_traj() 
        Trajectory = get_frame(Trajectory,options.frames) 
        tot_ions =  outcar.tot_ions  
        ion_list = outcar.ion_list 
        if fout != None: 
            traj_out = open(fout,'w') 
            for i in range(Trajectory.shape[0]): 
                traj_out.write('{} \nFrame {} \n'.format(tot_ions,i+1))
                for j in range(tot_ions): 
                    STR = ' {:<6}{:>12}{:>12}{:>12}'.format(ion_list[j] , Trajectory[i,j,0], Trajectory[i,j,1], Trajectory[i,j,2]) 
                    traj_out.write(STR+'\n')
                traj_out.write('\n') 
            traj_out.close()
        else: 
            print(Trajectory)
            
    elif options.job == 'abs_gaus': 
        if options.fname: 
            fname  = options.fname 
        else: 
            print(parser.print_help()) 
            print('Provide gaussian TDDFT output file name') 
            sys.exit () 

        if options.fout: 
            fout = options.fout 
        else: 
            fout = os.path.splitext(os.path.basename(options.fname))[0]
        
        unit = options.unit 
        
        flog = open(fname)
        lines = flog.read() 
        Excited_state = re.findall(r'Excited State.*',lines) 
        Excited_state =np.array([[i.split()[2].strip(':'),i.split()[4],i.split()[6],i.split()[8].split('=')[-1]] for i in Excited_state]).astype(float) 

        if options.unit=='ev': 
            X,OS = Excited_state[:,1],Excited_state[:,-1] 
        if options.unit=='nm': 
            X,OS = Excited_state[:,2],Excited_state[:,-1] 

        XX,YY = Dress_abs(X,OS,linewidth=options.sigma, unit=options.unit) 
        np.savetxt('Absorption_%s_lw%s.dat'%(fout,options.sigma), np.c_[XX,YY], fmt='%0.4f  %0.4f' if unit=='ev' else '%0.2f  %0.4f') 

        if options.xlim: 
            print('User provided xlim', options.xlim) 
            x1,x2 = options.xlim 
        else:
            if unit=='nm': 
                x1 = 250
                x2 = 50*(XX[YY>YY.max()*0.001][0]//50+1) 
            else: 
                x1 = 0.0
                x2 = 1+int(XX[YY>YY.max()*0.001][-1]) 
        if not  all([abs(x1-x2)<50, unit=='ev']) | all([abs(x1-x2)>50, unit=='nm']):
            print('\n**Unit of absorption and xlim are not compatible**\n') 
            
        y2 = YY[(XX>x1) & (XX<=x2)].max()*1.2   

        ylabel = r'Absorption, $Lmol^{-1}cm^{-1}$'
        xlabel = 'Energy, %s' %('eV' if unit=='ev' else unit) 

        plt.rcParams.update({'font.size': 20})
        
        plot_line(XX,YY, xlim=[x1,x2], ylim=[0,y2], xlabel=xlabel, ylabel=ylabel, figsize=options.figsize, 
                  name='Absorption_%s_lw%s.png'%(fout,options.sigma))  

        
                    
        #    print(a.outcar2incar())
        #    a.md_traj(out='Temmp_md.dat')  
        # F
    

if __name__ == "__main__":
    main() 
    
    
    



import numpy as np 
import re 
import sys 
import warnings


class read_outcar:
    def __init__(self, filename='OUTCAR') :  
        self.filename = filename
    
        self.charge = {}
        self.energies = None
        self.last_energy = None
        self.forces = None
        self.fermi = None
        self.bands = None
        self.positions = None
        self.kpoints = None
        self.array_sizes = {}
        self.species = None
        self.final_data = {}
        self.iteration_data = []
        self.outcar = open(self.filename, 'r').read() 
        
        
        self.poscar = None 
        self.paw_ions = []      # Ions from the potential function order in Outcar 
        self.no_ions = []  # Number of ions per ion type 
        self.paw_rwigs_rad = [] #RWIGS from postcar informatio in outcar 
        self.ibrion = None 
        self.incar = {} 
        #self.outcar_distance_warning()
        self.outcar_parser() 
        
        
    def outcar_parser(self): 
        f = self.outcar 
        #Reading poscar name 
        poscar = re.findall('POSCAR.*=.*',f)[-1].split('=') [-1].strip() 
        self.poscar = poscar  
        
        # reading the IONS from POTCAR information in outcar file 
        paw_ = re.findall('TITEL.*',f)
        PAW_ = [i.split('=')[1].split()[0] for i in paw_] 
        PAW_ions =  [i.split('=')[1].split()[1] for i in paw_] 
        self.paw_ions = PAW_ions 
        
        no_ions = [int(i) for i in re.findall('ions per type.*',f)[0].split('=')[-1].split()]
        self.no_ions = no_ions 
        
        paw_rwigs_rad = [float(i.split('=')[-1].split()[0]) for i in re.findall('RWIGS.*RWIGS.*',f) ] 
        self.paw_rwigs_rad = paw_rwigs_rad          
        
        ibrion = int(re.findall(r'IBRION.*',f)[0].split('=')[1].split()[0]) 
        self.ibrion = ibrion
        
        tot_ions = int(re.findall(r' number of ions     NIONS.*',f)[0].split('=')[1].split()[0])
        self.tot_ions = tot_ions 
        
        nkpoints = int(re.findall(r'NKPTS .*',f)[0].split('=')[1].split()[0]) 
        self.nkpoints = nkpoints 
        
        self.n_iteration = int(re.findall(r"Iteration   .*", f)[-1][10:16]) 
        self.nbands = int(re.findall(r"NBANDS.*", f)[0].split('=')[1].split()[0] ) 
        ion_list = [] 
        for i,j in enumerate(no_ions):
            ion_list = ion_list+[PAW_ions[i]]*j  
        self.ion_list = ion_list   
        
    def outcar2incar(self): 
        f = self.outcar 
        Tags = ['PREC','ISTART','ICHARG','ISPIN','LSORBIT','METAGGA','ENCUT','ENINI','NELM','EDIFF',
                'LREAL','EDIFFG','NSW','IBRION','ISIF','POTIM','TEIN','TEBEG','TEEND','SMASS','RWIGS',
                'EMIN','EFERMI','ISMEAR','IALGO','LWAVE','LCHARG','LORBIT','LMONO','LDIPOL','GGA','AEXX','NBANDS'] 
        Incar_var= {} 
        for var in Tags: 
            tag_value_  = re.findall(fr"{var}.*", f)[0].split('=')[1].split()[0] 
            Incar_var[var] = tag_value_
        self.INCAR = Incar_var 
        return Incar_var        
            
    def outcar_distance_warning(self) :                
        # Checking bond distances if unusual bond length in input geometry 
        # Getting the ions coordinates from outcar 
        f =  open(self.filename,'r')
        line = f.readline() 
        ion_pos = []
        ion_distance = [] 
        ion_distancewith = [] 
        while line: 
            if 'nearest neighbor table' in line: 
                line = f.readline() 
                while line:
                    try: 
                        kk,ll = line.split('-') 
                        n_ion = int(kk.split()[0]) 
                                                 
                        ion_pos.append([float(i) for i in kk.split()[-3:]]) 
                        k = int(len(ll.split())/2) 
                        ion_distance.append([float(ll.split()[1+i*2]) for i in range(k)])
                        ion_distancewith.append([int(ll.split()[i*2]) for i in range(k)])
                          #  print(line)
                        line = f.readline() 
                        if n_ion == sum(self.no_ions) : 
                            line = False  
                            break 
                    except ValueError as error: 
                        print(error.args[0])
                        if ('base 10' in error.args[0]) or ('to float' in error.args[0]): 
                            print(line)
                            print(error)
                            warnings.warn("Ions distance check: Distance can't convert to float in above line in OUTCAR file!" )
                            line = False 
                        else: 
                            line = f.readline() 
                            
                    except IndexError:  
                        line = f.readline() 
                break 
            else : 
                line = f.readline()            
        
        Corr_geom=True 
        for i,j in enumerate(ion_pos): 
            if any([l<0.10 for l in ion_distance[i]]): 
                print('Following ions are found in too close distance, \n check the input geometry') 
                print('Atom', j, ' and ', ion_distancewith[i], ' distance' , ion_distance[i] ) 
                Corr_geom=False 
                f.close() 
                sys.exit() 
        if Corr_geom==True: 
            print('No error found in initial coordinates')
        #f.close() 
    
    def input_geom(self, out=False,fout=None, **kwargs):  
        if fout==None: 
            fout = '%s_ini.xyz'%self.poscar
        f =  self.outcar   
        var =  "position of ions in fractional coordinates.*\n((.*\n){%i})"%self.tot_ions 
        frac_coordinates = re.findall(fr"{var}", f)[0][0].split('\n')  
        frac_coordinates = [i for i in frac_coordinates if len(i.strip()) != 0] 
                
        var =  "position of ions in cartesian coordinates.*\n((.*\n){%i})"%self.tot_ions 
        cart_coordinates = re.findall(fr"{var}", f)[0][0].split('\n')  
        cart_coordinates = [i for i in cart_coordinates if len(i.strip()) != 0]    

        if (len(kwargs)!=0 ) & ('coord' not in kwargs): 
            print('use kwargs coord=farc or cart to print coordinates')  
            return None 
        elif len(kwargs)==0: 
            kwargs['coord'] = 'cart' 
        if  kwargs['coord'].lower()[:4] == 'frac' :  
            CC = frac_coordinates 
        else : 
            CC = cart_coordinates 
        if len(CC) != self.tot_ions: 
            print("Fractional coordinated can't read properly, \n please check the output file fomrat")
            return None  
        else : 
            CC = ['{:<3}'.format(self.ion_list[i])+CC[i] for i in range(self.tot_ions)] 
            if out is True:  # Destination: user output, Flase : return a variable for other function 
                ff = open(fout,'w')
                ff.write(' %s\n %s \n'%(len(CC),self.poscar)) 
                for line in CC: 
                    ff.write('{:<3}{:>16}{:>16}{:>16} \n'.format(line.split()[0],line.split()[1],line.split()[2],line.split()[3]))
#                    ff.write(line+'\n') 
            else: 
#            print('Following is the input geometry in fractional coordinates') 
                return CC
        
    def md_traj(self,out=False, fout=None):
        if fout==None: 
            fout = '%s_traj.xyz'%self.poscar 
        f = self.outcar 
        n_iteration = self.n_iteration  # = int(re.findall(r"Iteration   .*", f)[-1][10:16])
        Trajectory = np.zeros((n_iteration, self.tot_ions,3))
        # Regex itreation is quite slow, need to change it 
        var =  "POSITION                                       TOTAL-FORCE.*\n.*\n((.*\n){%i})"%self.tot_ions
        geom_i = re.findall(fr"{var}", f)
#        geom_i = re.findall(r'POSITION                                       TOTAL-FORCE.*\n.*\n((.*\n){124})',f) 
        for j,i in enumerate(geom_i): 
            geom = np.loadtxt(i[0].split('\n')[:-1])[:,:3] 
            Trajectory[j,:,:] = geom          
        if len(geom_i) != n_iteration: 
            print('All ionic iteractions aren\'t finished completely')
        # Writing the file in xyz format 
        #print(out)
        if out != False: 
            traj_out = open(fout,'w') 
            for i in range(n_iteration): 
                traj_out.write('{} \nFrame {} \n'.format(self.tot_ions,i+1))
                for j in range(self.tot_ions): 
                    STR = ' {:<6}{:>12}{:>12}{:>12}'.format(self.ion_list[j] , Trajectory[i,j,0], Trajectory[i,j,1], Trajectory[i,j,2]) 
                    traj_out.write(STR+'\n')
                traj_out.write('\n') 
            traj_out.close()  
        else: 
            return Trajectory 
    
    def kpoint_dimension(self): 
        f=self.outcar 
        var = "k-points in units of 2pi.*\n((?:.*\n){%i})"%self.nkpoints 
        kpoints_dim_2pi = np.loadtxt( re.findall(fr"{var}", f).split('\n'),dtype=float)
        var = "k-points in reciprocal lattice and weights.*\n((?:.*\n){%i})"%self.nkpoints 
        kpoints_dim_recipr = np.loadtxt( re.findall(fr"{var}", f).split('\n'),dtype=float)  
        return kpoints_dim_2pi, kpoints_dim_recipr 
        
    def fermi_energies(self): 
        f=self.outcar
        fermi_all = re.findall(r'E-fermi : .*',f)  
        fermi_alphabeta = np.array([[float(i.split(':')[1].split()[0]),float(i.split(':')[3])] for i in fermi_all])
        return fermi_alphabeta 
    
    def band_energies(self) :
        f=self.outcar 
        States = np.zeros((self.nkpoints, self.n_iteration, self.nbands,3))
        
        for k in range(self.nkpoints): 
            var =  "k-point     %i.*\n.*band No.  band energies     occupation .*\n((?:.*\n){%i})"%(k+1,self.nbands) 
            bands  = re.findall(fr"{var}", f) 
            bands = np.array([np.loadtxt(i.split('\n')[:self.nbands] ) for i in bands ] )
            States[k,:,:,:] = bands 
#            bands = np.array([np.loadtxt(i[0].split('\n')[:nbands] ) for i in bands ] ) 
      #  j.split() for i in bands for j in i[0].split('\n')[:nbands]  
        return States 
       
    def free_energy(self): 
        f=self.outcar 
        free_en = np.array([i.split() for i in re.findall(r'free  energy.*', f)])[:,-2].astype(float) 
        return free_en 

class read_procar: 
    def __init__(self, filename='PROCAR') :  
        self.filename = filename
    

            
#if __name__ == '__main__': 
#    f = 'Datafiles/OUTCAR' 
#    a = Output(filename=f)
#    print(a.outcar2incar())
#    a.md_traj(out='Temmp_md.dat')  
#  'NKPTS', 'NBANDS', 'NEDOS', 'NIONS', 'NGX', 'NGY', 'NGZ', 'NGXF', 'NGYF', 'NGZF', 'ISPIN']: 


# -*- coding: utf-8 -*-
"""
Created on Fri Dec 23 21:33:16 2022

@author: Mohammed a jabed, jabed.abu@gmail.com
"""
import numpy as np 
import math 
import os 

def angle(a,b):
    import numpy as np 
    a=np.asarray(a) 
    b=np.asarray(b) 
    unitV1 = a/np.linalg.norm(a)
    unitV2 = b/np.linalg.norm(b)
    return  np.arccos(np.dot(unitV1, unitV2))


def contcar2com(name, name_out,Gaus_out=False):  # name is the contcar file name 
    f_cont = open(name,'r') 
    # reading file, and cell normalizing factor  
    line_cont = f_cont.readlines()  
    sys_name  = line_cont[0] #compname  
    sigma = float(line_cont[1].split()[0]) 

    #Checking file format, cell dimension check 
    if not all([len(i.split())==3 for i in line_cont[2:5]]): 
        print(line_cont[2:5])
        print('File format does\'t match with CONTCAR file format, script is quiting')
        exit() 

        #Reading unit cell dimention 
    xyz = np.loadtxt(line_cont[2:5], dtype=float)
    print('The unit cell size is: ')
    print(xyz)

    Elements = line_cont[5].split() 	#Elements list 
    El_num = [int(i) for i in line_cont[6].split()]    #number of atoms of each elements 
    if line_cont[7].strip()[0].lower()=='s': 
        selec_dynm = True
        if line_cont[8].strip()[0].lower()=='d': 
            FracCoord=True 
        elif line_cont[8].strip()[0].lower()=='c':
            FracCoord=False
    elif line_cont[7].strip()[0].lower()=='d': 
        FracCoord=True
        selec_dynm = False
    elif line_cont[7].strip()[0].lower()=='c':
        FracCoord=False
        selec_dynm = False
    else: 
        "Can\'t read cartesian or direct coordinate correctly, line number 8 and 9. \nScript is quiting now!! :( "
        exit() 

    if selec_dynm is True:  
        Cord=np.loadtxt(line_cont[9:9+sum(El_num)], dtype=str)[:,:3].astype(float)
        AtomFrez = (np.loadtxt(line_cont[9:9+sum(El_num)], dtype=str)[:,3:]=='F').all(axis=1)
        Freez =  AtomFrez*-1
    else:
        try: 
            Cord = np.loadtxt(line_cont[8:8+sum(El_num)], dtype=str).astype(float) 
        except ValueError:  
            print('Trying to read Direct coordinate format, found string in line, \nScript is quiting') 
            exit() 

 
    ions_list = [] # Atom_symbol =[] 
    for i in range(len(Elements)): 
        ions_list = ions_list + [Elements[i]]*El_num[i] 

    #Convertion matices 
    a1,a2,a3 = xyz[0,:]
    b1,b2,b3 = xyz[1,:]
    c1,c2,c3 = xyz[2,:] 
    #Length of the Parallelepiped edge 
    a,b,c= np.linalg.norm(xyz[0,:]), np.linalg.norm(xyz[1,:]), np.linalg.norm(xyz[2,:])
    # angel of Parallelepiped 
    alpha = angle([b1,b2,b3],[c1,c2,c3])  
    beta  = angle([a1,a2,a3],[c1,c2,c3]) 
    gamma  = angle([a1,a2,a3],[b1,b2,b3]) 
    #Simplified term. look on the Wikipedia for more 
    n = (math.cos(alpha)-math.cos(gamma)*math.cos(beta))/math.sin(gamma) 
    Trans_matrix_inv = np.array([[a,  b*math.cos(gamma), c*math.cos(beta),],
                                 [0,  b*math.sin(gamma), c*n],
                                 [0,  0,                 c*((math.sin(beta))**2-n**2)**0.5]])  

    if FracCoord is True:
        Cart_Cord = np.matmul(Trans_matrix_inv,Cord.T).T
    elif FracCoord is False:
        Cart_Cord = Cord 

    if selec_dynm is True: 
        AA = np.c_[np.asarray(ions_list),Freez]  
        Cord_Str = np.c_[AA,np.asarray(Cart_Cord).round(4)] 
       # print(np.asarray(Cart_Cord).round(4))
    elif selec_dynm is False: 
        Cord_Str = np.c_[np.asarray(ions_list),np.asarray(Cart_Cord).round(4)] 
      #  Cord_Str = np.c_[np.asarray(ions_list),np.asarray(np.array_str(np.asarray(Cart_Cord),precision=4))] 
   
    print(name_out)
    if Gaus_out == True: 
        Header = '''%%mem=20GB
%%nprocshared=40
%%chk=%s.chk 

#p opt pbe1pbe/gen nosymm pseudo=read scf=maxcycles=10000

%s

0 1
''' %(name_out,sys_name.rstrip('\n')) 

        with open('%s.com'%name_out, "w") as fout:
            fout.write(Header)
   #         print(Cord_Str)
            if selec_dynm is True: 
                np.savetxt(fout, Cord_Str,fmt='%-5.10s %-5.10s %-10.10s  %-10.10s  %-10.10s') 
                
            elif selec_dynm is False: 
                np.savetxt(fout, Cord_Str,fmt='%-5.10s  %-10.6s  %-10.10s  %-10.10s')
               
            PBC = np.c_[np.asarray(['Tv','Tv','Tv']),xyz.astype('<U10')] 
            np.savetxt(fout, PBC,fmt='%-5.10s  %-010.8s  %-10.8s  %-10.10s') 
             
        print('Gaussian input file is written in the file name %s.com' %name_out)  
        fout.close() 


    else :
        with open('%s.xyz'%name_out.rstrip('.xyz'), "a") as fout: 
            #print(len(Cord_Str))
            fout.write('%i \n' %len(Cord_Str))
            fout.write('%s \n' %sys_name.rstrip('\n')) 
            np.savetxt(fout,np.c_[Cord_Str[:,0],Cord_Str[:,-3:]],
                       fmt='%-5.10s  %-10.10s  %-10.10s  %-10.10s') 
        print('xyz file is written in the file name %s.xyz' %name_out) 
        fout.close() 

def com2poscar(name,name_out=True, select=False, direct=True): 
    def angle(a,b):
        a=np.asarray(a) 
        b=np.asarray(b) 
        unitV1 = a/np.linalg.norm(a)
        unitV2 = b/np.linalg.norm(b)
        return  np.arccos(np.dot(unitV1, unitV2))
    
    file = open(name) 
    line = file.readline() 
    topheader = [] 
    commandkeys = [] 
     
    while line : 
        print('%' in line)
        if '%' in line: 
#            print(line)
            topheader.append(line)
            line = file.readline() 
#            print('command')
        else: 
            break 
    if len(line.strip())==0: 
        line = file.readline() 
    while line : 
        if '#' in line: 
            commandkeys.append(line)
            line=file.readline() 
#            print('keys')
        else: 
            break 
    if len(line.strip())==0: 
        line=file.readline() 
        
    comp_name = line.strip('\n') 
    line = file.readline() 
#    print('name')
        
    if len(line.strip())==0: 
        line=file.readline()  
    Coord = []
    TVs= [] 
    while line: 
        if (len(line.split())%2 ==0 ) & all([i.isdigit() for i in line.split()]) :
#            print('charg_multi')
            chrg_multi = line.strip('\n') 
            line = file.readline() 
            while line :
                if not 'tv' in line.lower(): 
                    Coord.append(line.strip('\n') )
                else: 
                    TVs.append(line.strip('\n').split()) 
                line = file.readline()
                if len(line.strip())==0: 
                    break
            break 
        
    # Warning 
    if len(TVs) <3: 
        print('Periodic cell dimensions not found.') 
        return None  
    
    freeztag = lambda a : False if a==-1 else True 
    ionsFrag = lambda a : a.split('(')[0].strip() if 'frag' in a.lower() else a.strip() 
    
    coordinates_xyz = np.empty((len(Coord),5), dtype=object) 
    for j,coord_ in enumerate(Coord):
        ll = coord_.split() 
        coordinates_xyz[j,0]  = ionsFrag( ll[0])   
        coordinates_xyz[j,2:5] = ll[-3:] 
        if len(ll) ==4: 
            coordinates_xyz[j,1] = True  
        elif len(ll) ==5: 
            coordinates_xyz[j,1] = freeztag(int(ll[1]))  
   
    coordinates_xyz = coordinates_xyz[coordinates_xyz[:,0].argsort()] 
    
    
    
    #working on Cartesian Fractional coordinates conversion 
    print(np.asarray(TVs)) 
    TVV = np.asarray(TVs)[:,1:].astype(float) 
    a1,a2,a3 = TVV[0,:]
    b1,b2,b3 = TVV[1,:]
    c1,c2,c3 = TVV[2,:] 
    #Length of the Parallelepiped edge 
    a,b,c= np.linalg.norm(TVV[0,:]), np.linalg.norm(TVV[1,:]), np.linalg.norm(TVV[2,:])
    # angel of Parallelepiped 
    alpha = angle([b1,b2,b3],[c1,c2,c3])  
    beta  = angle([a1,a2,a3],[c1,c2,c3]) 
    gamma  = angle([a1,a2,a3],[b1,b2,b3]) 
    #Simplified term. look on the Wikipedia for more 
    n = (math.cos(alpha)-math.cos(gamma)*math.cos(beta))/math.sin(gamma) 
    Trans_matrix_inv = np.array([[a,  b*math.cos(gamma), c*math.cos(beta),],
                                 [0,  b*math.sin(gamma), c*n],
                                 [0,  0,                 c*((math.sin(beta))**2-n**2)**0.5]]) 
    Trans_matrix = np.linalg.inv(Trans_matrix_inv)
#    delta = a*b*c*(1-math.cos(alpha)**2-math.cos(beta)**2-math.cos(gamma)**2+
 #                  2*math.cos(alpha)*math.cos(beta)*math.cos(gamma))**0.5 
    #Convertion matices 
#    Trans_matrix = np.array([[1/a, -1/(math.tan(gamma)*a), b*c*(math.cos(alpha)*math.cos(gamma)-math.cos(beta))/(delta*math.sin(gamma)) ],
#                          [0, 1/(b*math.sin(gamma)), a*c*(math.cos(beta)*math.cos(gamma) - math.cos(alpha))/(delta*math.sin(gamma))],
#                          [0,0, a*b*math.sin(gamma)/delta]
 #                         ])
     
    frac_coord = np.matmul(Trans_matrix,coordinates_xyz[:,2:].astype(float).T).T 
    ions_name = list(set(coordinates_xyz[:,0]))
    ions_name.sort() 
    print(ions_name)
    ions_count = [] 
    for j in ions_name: 
        ions_count.append((coordinates_xyz[:,0] == j).sum())

    if name_out==None:
        print('Parallelepiped vectors are - ') 
        print(TVV)
        print('Ions are', ions_name) 
        print('Ions counts are: ',ions_count) 
        return frac_coord
    else: 
        name_ = os.path.splitext(name)[0] 
        name_ = os.path.basename(name_)
        if name_out == True:            
            fout = open('{}.POSCAR'.format(name_),'w') 
        else: 
            fout = open('{}'.format(name_out),'w')  
        
        fout.write(name_+'\n')
        fout.write(' 1.000000  \n') 
        for j in range(3): 
            fout.write('  %0.6f   %0.6f    %0.6f  \n' %(TVV[j,0],TVV[j,1],TVV[j,2])) 
        
        # Ions types 
        for i in ions_name: 
            fout.write(' %s '%i) 
        fout.write('\n') 
        #Number of each Ions type 
        for i in ions_count: 
            fout.write(' %i ' %i) 
        fout.write('\n') 
        
        #Writing the Coordinates
        if select: 
            fout.write('select \n') 
        if direct: 
            fout.write('Direct \n') 
        else: 
            fout.write('Cartesian \n') 
        
        for j in range(coordinates_xyz.shape[0]): 
            if direct == False: 
                new_xyz = np.asarray(coordinates_xyz[j,2:],dtype=float ) 
            else: 
                new_xyz = frac_coord[j,:] 
                
            if select:
                if coordinates_xyz[j,1] == True:
                    sel_str = '  T      T       T \n'
                elif coordinates_xyz[j,1] == False: 
                    sel_str = '  F      F       F  \n'
            else:
                sel_str = '  \n'
#            fout.write('%0.6f   %0.6f   %0.6f   %s' %(new_xyz[0],new_xyz[1],new_xyz[2],sel_str)) 
            fout.write('{:> 12.7f}   {:> 12.7f}   {:> 12.7f}   {}'.format(new_xyz[0],new_xyz[1],new_xyz[2],sel_str))
        print('POSCAR file written in the file %s' %name_out)
        fout.close() 
            



